import re
import pickle
from math import sqrt
from pathlib import Path
from collections import defaultdict, Counter
from datetime import datetime, timedelta
import pandas as pd

import tqdm
import torch
import pytorch_lightning as pl
import matplotlib.pyplot as plt


class AnomalyData(torch.utils.data.Dataset):
    def __init__(self, filename: str, dataname: str, window_size):
        self.df = pd.read_excel(filename)[dataname]
        self.mini = self.df.min()
        self.maxi = self.df.max()
        self.data = self.make_window(list(self.df), window_size=window_size)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        return self.data[idx]

    def make_window(self, inp: list[float], window_size: int, is_y=False):
        """연속된 데이터를 window size만큼 끊어서 개별 데이터로 만듭니다.
        Args
            inp: 전체 데이터
            window_size: 결과 도출에 이전 몇 개의 데이터를 참조할 것인가
        """
        ret = []
        for i in range(len(inp)-window_size):
            if is_y:
                # label인 경우 1 time-step 이후의 데이터를 리턴
                x = inp[i+window_size]
            else:
                x = inp[i:i+window_size]
            ret.append(x)
        return self.normalize(torch.tensor(ret).unsqueeze(2))

    def normalize(self, inp: torch.Tensor):
        """데이터를 -1 ~ 1 사이의 값으로 normalize합니다."""
        return (inp - self.mini) / (self.maxi - self.mini) * 2 - 1

    def denormalize(self, inp: torch.Tensor):
        """데이터를 -1 ~ 1 사이의 값에서 원래 값으로 denormalize합니다."""
        return ((inp + 1) / 2) * (self.maxi - self.mini) + self.mini


class AnomalyLSTM(pl.LightningModule):
    def __init__(self, hparams: dict = None):
        super().__init__()
        if hparams is None:
            hparams = {}
        self.hparams.update(hparams)
        self.threshold = self.hparams["threshold"]
        self.lstm_encoder = torch.nn.LSTM(
            input_size=hparams["input_size"],
            hidden_size=128,
            batch_first=True,
        )
        self.lstm_decoder = torch.nn.LSTM(
            input_size=128,
            hidden_size=hparams["input_size"],
            batch_first=True,
        )
        self.mse = torch.nn.MSELoss(reduction='none')

    def forward(self, inp):
        """
        input:
            [B, ]
        return:
            [B, ]
        """
        latent_vector, _ = self.lstm_encoder(inp)
        x_prime, _ = self.lstm_decoder(latent_vector)
        return x_prime

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=float(self.hparams['lr']))
        return optimizer

    def training_step(self, batch, batch_nb):
        out = self(batch)
        anomaly_score = self.cal_loss(batch, out)

        self.log('train_loss', anomaly_score.mean(), batch_size=len(batch))
        return {'loss': anomaly_score.mean()}

    def validation_step(self, batch, batch_nb):
        out = self(batch)
        anomaly_score = self.cal_loss(batch, out)

        self.log('val_loss', anomaly_score.mean(), batch_size=len(batch))
        return {'loss': anomaly_score.mean()}

    def predict_step(self, seq, batch_idx: int, dataloader_idx: int = 0):
        out = self(seq)
        losses = self.cal_loss(out, seq)
        return losses.tolist()

    def cal_loss(self, pred, seq):
        """
        Define how to calculate loss

        logits:
            [B, ]
        targets:
            [B, ]
        """
        losses = self.mse(pred, seq).mean(dim=1).squeeze()
        return losses

if __name__ == "__main__":
    hparams = {
        "input_size": 1,
        "seq_len": 16,
        "threshold": 0.01,  # 모델 결과에서 anomaly로 간주할 문턱값
        "tolerance": 3, # threshold가 넘는 값이 몇번 연속으로 출현해야 anomaly로 간주할 지
        "epochs": 10,
        "lr": 0.0001,
        "batch_size": 128,
        "datafile": "303-bed.xlsx",
        "dataname": "voc",  # xlsx 파일의 열 이름을 입력 (temp, humi, co2, ...)
    }
    print(hparams)

    dataset = AnomalyData(
        hparams["datafile"],
        dataname=hparams["dataname"],
        window_size=hparams["seq_len"]
    )
    train_dataset, valid_dataset = torch.utils.data.random_split(
        dataset, lengths=[int(len(dataset)*0.9), len(dataset)-int(len(dataset)*0.9)]
    )

    train_dataloader = torch.utils.data.DataLoader(train_dataset,
                                                   batch_size=hparams["batch_size"],
                                                   shuffle=True,
                                                   num_workers=1)
    valid_dataloader = torch.utils.data.DataLoader(valid_dataset,
                                                   batch_size=hparams["batch_size"],
                                                   num_workers=1)
    predict_dataloader = torch.utils.data.DataLoader(dataset,
                                                     batch_size=hparams["batch_size"],
                                                     num_workers=1)

    model = AnomalyLSTM(hparams)
    device_cnt = torch.cuda.device_count()
    trainer = pl.Trainer(devices=1,
                         accelerator="cpu",
                         max_epochs=hparams["epochs"])
    trainer.fit(model,
                train_dataloaders=train_dataloader,
                val_dataloaders=valid_dataloader)

    # get predictions
    pred = [0] * (hparams["seq_len"] - 1)
    for x in trainer.predict(model, dataloaders=predict_dataloader):
        pred.extend(x)

    # anomaly positive 구간 확인
    positives = []
    period = []
    for i, x in enumerate(pred):
        if x >= hparams["threshold"] and not period:
            period.append(i)
        elif x < hparams["threshold"] and period:
            period.append(i)
            if period[1] - period[0] > hparams["tolerance"]:
                positives.append(period)
            period = []

    # 그래프 그리기
    fig, (ax1, ax2) = plt.subplots(2, 1, sharey=False, figsize=(16, 6), layout="constrained")
    ax1.plot(
        pd.read_excel(hparams["datafile"])[hparams["dataname"]],
        label=hparams["dataname"]
    )
    for (start, end) in positives:
        ax1.axvspan(start, end, color="red", alpha=0.3)
    ax1.set_title(hparams["dataname"])
    ax2.plot(pred, label="anomaly score")
    ax2.set_title("anomaly score")
    plt.show()
